#include "types.h"
#include "libc.h"
#include "sim4.h"
#include "psublast.h"
#include "splice.h"
#include "sites.h"
#include "glimmerSplice.h"

#ifndef __lint
/*@unused@*/
static const char rcsid[] =
"$Id: splice.c,v 1.2 2008/07/19 02:48:23 florea Exp $";
#endif


#define GENESPLICER_SPAN    80
#define GLIMMER_XSPAN       30
#define GLIMMER_ISPAN       20
#define S4_SPAN              0
#define MAX_SPAN            80

static char  Glimmer_TRAIN_DIR[] = "./GlimmerModels/";
static char  Glimmer_posDonModelPath[] = "donors.162.pos.icm";
static char  Glimmer_negDonModelPath[] = "donors.162.neg.icm";
static char  Glimmer_posAccModelPath[] = "acceptors.162.pos.icm";
static char  Glimmer_negAccModelPath[] = "acceptors.162.neg.icm";

static int   spl_encode[NACHARS], rev_compl[NACHARS];
static char *nsegmentL, *nsegmentR, *asegmentL, *asegmentR;
static int   spliceInit = 0;


// Move to glimmer.h
struct  Fixed_Length_ICM_t  donor_pos_model, donor_neg_model;
struct  Fixed_Length_ICM_t  acceptor_pos_model, acceptor_neg_model;
int     donor_pos_model_len, donor_neg_model_len;
int     acceptor_pos_model_len, acceptor_neg_model_len;
int     initGlimmerModel = 0;


signal_t gt = {{0, 0, 0, 2, 0},{0, 0, 0, 2, 0},{2, 3, 2, 5, 2},{0, 0, 0, 2, 0},{0, 0, 0, 2, 0}};
signal_t ct = {{0, 0, 0, 2, 0},{2, 2, 2, 5, 2},{0, 0, 0, 2, 0},{0, 0, 0, 2, 0},{0, 0, 0, 2, 0}};
signal_t ag = {{2, 2, 5, 2, 2},{0, 0, 2, 0, 0},{0, 0, 2, 0, 0},{0, 0, 2, 0, 0},{0, 0, 2, 0, 0}};
signal_t ac = {{2, 5, 2, 2, 2},{0, 2, 0, 0, 0},{0, 3, 0, 0, 0},{0, 2, 0, 0, 0},{0, 2, 0, 0, 0}};

static void splice_original(uchar *,int,int,int,int,uchar *,int,int,double *,double *,double *,double *,int);
static void splice_GeneSplicer(uchar *,int,int,int,int,uchar *,int,int,double *,double *,double *,double *,int);
static void splice_Glimmer(uchar *,int,int,int,int,uchar *,int,int,double *,double *,double *,double *,int);

static void loadGlimmerModel (char *train_dir) 
{
   char filename[1000];

   sprintf(filename, "%s/%s", train_dir, Glimmer_posDonModelPath);
   readModel (&donor_pos_model, filename);

   sprintf(filename, "%s/%s", train_dir, Glimmer_negDonModelPath);
   readModel (&donor_neg_model, filename);

   sprintf(filename, "%s/%s", train_dir, Glimmer_posAccModelPath);
   readModel (&acceptor_pos_model, filename);

   sprintf(filename, "%s/%s", train_dir, Glimmer_negAccModelPath);
   readModel (&acceptor_neg_model, filename);

   donor_pos_model_len = getModelLength (donor_pos_model);
   donor_neg_model_len = getModelLength (donor_neg_model);
   acceptor_pos_model_len = getModelLength (acceptor_pos_model);
   acceptor_neg_model_len = getModelLength (acceptor_neg_model);
      
   if (donor_pos_model_len!=donor_neg_model_len)
      fatal ("ERROR:  Positive and negative donor model lengths differ\n");
   if (acceptor_pos_model_len!=acceptor_neg_model_len)
      fatal ("ERROR:  Positive and negative acceptor model lengths differ\n");

   return;
}

static double  ScoreDonor_Glimmer (char *asegment, char *train_dir)
{
   double  pos_score, neg_score, diff;

   if (!initGlimmerModel) {
      loadGlimmerModel(train_dir);
      initGlimmerModel = 1;
   }

   pos_score = Score_Window (donor_pos_model, asegment, GLIMMER_XSPAN);
   neg_score = Score_Window (donor_neg_model, asegment, GLIMMER_XSPAN);
   diff = pos_score - neg_score;

// printf ("%s  %9.5f  %9.5f  %9.5f\n", string, pos_score, neg_score, diff);

   return diff;
}

static double  ScoreAcceptor_Glimmer (char *asegment, char *train_dir)
{
   double  pos_score, neg_score, diff;

   if (!initGlimmerModel) {
      loadGlimmerModel(train_dir);
      initGlimmerModel = 1;
   }

   pos_score = Score_Window (acceptor_pos_model, asegment, GLIMMER_ISPAN);
   neg_score = Score_Window (acceptor_neg_model, asegment, GLIMMER_ISPAN);
   diff = pos_score - neg_score;

// printf ("%s  %9.5f  %9.5f  %9.5f\n", string, pos_score, neg_score, diff);

   return diff;
}

#ifdef DEBUG
static void print_splice(splice_t *);
#endif

static void splice_donor(uchar *xseq, uchar *yseq, int M, int N, double *gt_score,
                         double *ct_score, double **max_Gf, double **max_Cf,
                         double **start_Gi, double **start_Ci);
static void splice_donor_uni(uchar *xseq, uchar *yseq, int M, int N,
                         double *It_score, double **max_IF, double **end_Ii);
static void splice_acceptor(uchar *xseq, uchar *yseq, int M, int N,
                         double *ag_score, double *ac_score, double **max_Gb,  
                         double **max_Cb, double **end_Gi, double **end_Ci);
static void splice_acceptor_uni(uchar *xseq, uchar *yseq, int M, int N,
                         double *aI_score, double **max_Ib, double **end_Ii);
static int  stepct(int n);


void splice_init(int spl_model)
{
   int i;
                                                                                                                                 
   if (!spliceInit) {
       for (i=0; i<NACHARS; spl_encode[i]=0, rev_compl[i]='T', i++);
                                                                                                                                 
       spl_encode['A'] = 0; spl_encode['C'] = 1;
       spl_encode['G'] = 2; spl_encode['T'] = 3;
                                                                                                                                 
                                                                                                                                 
       rev_compl['A'] = 'T'; rev_compl['C'] = 'G';
       rev_compl['G'] = 'C'; rev_compl['T'] = 'A';
                                                                                                                                 
                                                                                                                                 
       nsegmentL = (char *) ckalloc(2*MAX_SPAN + 2*MAX_SLIDE + 3);
       nsegmentR = (char *) ckalloc(2*MAX_SPAN + 2*MAX_SLIDE + 3);
                                                                                                                                 
                                                                                                                                 
       /* Obs: for Glimmer scoring, need only remember the reverse complemented
          segments; but for now we allocate two arrays */
       if (spl_model==SPLICE_GLIMMER) {
          asegmentL = (char *) ckalloc(2*MAX_SPAN + 2*MAX_SLIDE + 3);
          asegmentR = (char *) ckalloc(2*MAX_SPAN + 2*MAX_SLIDE + 3);
       }
                                                                                                                                 
                                                                                                                                 
       spliceInit = 1;
   }
}
                                                                                                                                 
void splice(uchar *in_seqx, int ls, int us, int le, int ue,
            uchar *in_seqy, int ys, int ye,
            splice_t **gcell, splice_t **ccell, int ori, int spl_model)
{
   double *gtscore=NULL, *ctscore=NULL, *agscore=NULL, *acscore=NULL;
   int i, tmp;
   double maxCscore, maxGscore;
   int Gxs, Gxe, Gy, Cxs, Cxe, Cy;
   int keep_Ci, keep_Gi;
   double *max_Cf=NULL, *max_Gf=NULL, *max_Cb=NULL, *max_Gb=NULL;
   double *start_Gi=NULL, *start_Ci=NULL, *end_Gi=NULL, *end_Ci=NULL;

   splice_init(spl_model);
                                                                                                                                 
   if (ori==FWD || ori==BOTH) {
       gtscore = (double *)ckalloc(((us-ls+2)+(ue-le+2))*sizeof(double));
       agscore = gtscore+(us-ls+2);
   }
   if (ori==BWD || ori==BOTH) {
       ctscore = (double *)ckalloc(((us-ls+2)+(ue-le+2))*sizeof(double));
       acscore = ctscore+(us-ls+2);
   }

   switch (spl_model) {
      case SPLICE_ORIGINAL:
           splice_original(in_seqx,ls,us,le,ue,in_seqy,ys,ye,gtscore,agscore,ctscore,acscore,ori);
           break;

      case SPLICE_GENESPLICER:
           splice_GeneSplicer(in_seqx,ls,us,le,ue,in_seqy,ys,ye,gtscore,agscore,ctscore,acscore,ori);
           break;

      case SPLICE_GLIMMER:
           splice_Glimmer(in_seqx,ls,us,le,ue,in_seqy,ys,ye,gtscore,agscore,ctscore,acscore,ori);
           break;
   
      default: 
           fatalf("Unrecognized splice model. %d\n", spl_model);
   }

   if (ori==FWD) {
      splice_donor_uni(in_seqx+ls-1, in_seqy+ys-1, us-ls+1, ye-ys+1,
                       gtscore, &max_Gf, &start_Gi);
      splice_acceptor_uni(in_seqx+le-1, in_seqy+ys-1, ue-le+1, ye-ys+1,
                       agscore, &max_Gb, &end_Gi);
      free(gtscore); /* free(agscore); */
                                                                                                                                 
   } else if (ori==BWD) {
      splice_donor_uni(in_seqx+ls-1, in_seqy+ys-1, us-ls+1, ye-ys+1,
                       ctscore, &max_Cf, &start_Ci);
      splice_acceptor_uni(in_seqx+le-1, in_seqy+ys-1, ue-le+1, ye-ys+1,
                       acscore, &max_Cb, &end_Ci);
      free(ctscore); /* free(acscore); */
                                                                                                                                 
   } else {
      splice_donor(in_seqx+ls-1, in_seqy+ys-1, us-ls+1, ye-ys+1,
                   gtscore, ctscore, &max_Gf, &max_Cf, &start_Gi, &start_Ci);
      splice_acceptor(in_seqx+le-1, in_seqy+ys-1, ue-le+1, ye-ys+1,
                      agscore, acscore, &max_Gb, &max_Cb, &end_Gi, &end_Ci);
                                                                                                                                 
      free(gtscore); /* free(agscore);     */
      free(ctscore); /* free(acscore); */
   }
                                                                                                                                 
   maxCscore = -999999; maxGscore = -999999;
   Gxs = Gxe = Gy = Cxs = Cxe = Cy = -1;
   if (ori==FWD || ori==BOTH) {
      for (i=0; i<=ye-ys+1; i++) {
        if ((tmp=max_Gf[i]+max_Gb[i])>maxGscore) {
            maxGscore = tmp;
            /* save (i, start_Gi[i], end_Gi[i]); */
            Gxs = ls+start_Gi[i]-1; Gxe = le+end_Gi[i]-1; Gy = ys+i-1;
            keep_Gi = i;
        }
      }
      free(max_Gf); free(max_Gb); /* free(start_Gi); free(end_Gi); */
   }
   if (ori==BWD || ori==BOTH) {
      for (i=0; i<=ye-ys+1; i++) {
        if ((tmp=max_Cf[i]+max_Cb[i])>maxCscore) {
            maxCscore = tmp;
            /* save (i, start_Ci[i], end_Ci[i]); */
            Cxs = ls+start_Ci[i]-1; Cxe = le+end_Ci[i]-1; Cy = ys+i-1;
            keep_Ci = i;
        }
      }
      free(max_Cf); free(max_Cb); /* free(start_Ci); free(end_Ci); */
   }
                                                                                                                                                                                                                                                                   
   *gcell = new_splice('G', Gxs, Gxe, Gy, Gy+1, maxGscore, NULL);
   *ccell = new_splice('C', Cxs, Cxe, Cy, Cy+1, maxCscore, NULL);
                                                                                                                                                                                                                                                                   
#ifdef DEBUG
   print_splice(*gcell); print_splice(*ccell);
#endif

   return;
}

static void splice_original(uchar *in_seqx, int ls, int us, int le, int ue,
                            uchar *in_seqy, int ys, int ye,
                            double *gtscore, double *agscore,
                            double *ctscore, double *acscore, int ori)
{
   int p, q, i;
   uchar  *s,*t, ch;

   for (i=0, s=in_seqx+ls-MAX_SPAN-1; i<2*MAX_SPAN+us-ls+3; nsegmentL[i++] = spl_encode[*s++]);
   for (i=0, s=in_seqx+le-2-MAX_SPAN-1; i<2*MAX_SPAN+ue-le+3; nsegmentR[i++] = spl_encode[*s++]);
                                                                                                                                 
   if (ori==FWD || ori==BOTH) {

       for (p=0, s=nsegmentL+MAX_SPAN; p<=us-ls+1; p++, s++)
           gtscore[p] = gt[*s][*(s+1)];
       for (q=ue-le+1, s=nsegmentR+MAX_SPAN+ue-le+2; q>=0; q--, s--)
           agscore[q] = ag[*(s-1)][*s];
   }
                                                                                                                                 
   if (ori==BWD || ori==BOTH) {
                                                                                                                                 
       /* reverse complement the nsegments, 0-3 alphabet */
       for (s=nsegmentL, t=nsegmentL+2*MAX_SPAN+us-ls+3-1; s<t; s++, t--)
           { ch = 3-(*s); *s = 3-(*t); *t = ch; }
       for (s=nsegmentR, t=nsegmentR+2*MAX_SPAN+ue-le+3-1; s<t; s++, t--)
           { ch = 3-(*s); *s = 3-(*t); *t = ch; }
                                                                                                                                 
       for (p=0, s=nsegmentL+MAX_SPAN+us-ls+2; p<=us-ls+1; p++, s--)
           ctscore[p] = ag[*(s-1)][*s];
       for (q=ue-le+1, s=nsegmentR+MAX_SPAN; q>=0; q--, s++)
           acscore[q] = gt[*s][*(s+1)];
   }
                                                                                                                                 
   return;
}

static void splice_GeneSplicer(uchar *in_seqx, int ls, int us, int le, int ue,
                               uchar *in_seqy, int ys, int ye,
                               double *gtscore, double *agscore,
                               double *ctscore, double *acscore, int ori)
{
   int p, q, i;
   uchar  *s,*t, ch;
                                                                                                                                 
   for (i=0, s=in_seqx+ls-MAX_SPAN-1; i<2*MAX_SPAN+us-ls+3; nsegmentL[i++] = spl_encode[*s++]);
   for (i=0, s=in_seqx+le-2-MAX_SPAN-1; i<2*MAX_SPAN+ue-le+3; nsegmentR[i++] = spl_encode[*s++]);
                                                                                                                                 
   if (ori==FWD || ori==BOTH) {

       for (p=0, s=nsegmentL+MAX_SPAN; p<=us-ls+1; p++, s++) {
           gtscore[p] = ScoreDonor_GeneSplicer(s-GENESPLICER_SPAN);

           if (gtscore[p] < -14) gtscore[p] = -14.0;
           if (gtscore[p] > 19) gtscore[p] = 19;
           gtscore[p] = 5.0*(gtscore[p]+14.0)/33.0;
           gtscore[p] = 0.4*gtscore[p] + 0.6*gt[*s][*(s+1)];
       }
       for (q=ue-le+1, s=nsegmentR+MAX_SPAN+ue-le+2; q>=0; q--, s--) {
           agscore[q] = ScoreAcceptor_GeneSplicer(s-GENESPLICER_SPAN-1);

           if (agscore[q] < -23) agscore[q] = -23.0;
           if (agscore[q] > 20) agscore[q] = 20.0;
           agscore[q] = 5.0*(agscore[q]+23.0)/43.0;
           agscore[q] = 0.4*agscore[q] + 0.6*ag[*(s-1)][*s];
       }
   }
                                                                                                                                 
   if (ori==BWD || ori==BOTH) {

       /* reverse complement the nsegments, 0-3 alphabet */
       for (s=nsegmentL, t=nsegmentL+2*MAX_SPAN+us-ls+3-1; s<t; s++, t--)
           { ch = 3-(*s); *s = 3-(*t); *t = ch; }
       for (s=nsegmentR, t=nsegmentR+2*MAX_SPAN+ue-le+3-1; s<t; s++, t--)
           { ch = 3-(*s); *s = 3-(*t); *t = ch; }
                                                                                                                                 
       for (p=0, s=nsegmentL+MAX_SPAN+us-ls+2; p<=us-ls+1; p++, s--) {
           ctscore[p] = ScoreAcceptor_GeneSplicer(s-GENESPLICER_SPAN-1);
                                                                                                                                 
           if (ctscore[p] < -23) ctscore[p] = -23.0;
           if (ctscore[p] > 20) ctscore[p] = 20.0;
           ctscore[p] = 5.0*(ctscore[p]+23.0)/43.0;
           ctscore[p] = 0.4*ctscore[p] + 0.6*ag[*(s-1)][*s];
       }
       for (q=ue-le+1, s=nsegmentR+MAX_SPAN; q>=0; q--, s++) {
           acscore[q] = ScoreDonor_GeneSplicer(s-GENESPLICER_SPAN);
                                                                                                                                 
           if (acscore[q] < -14) acscore[q] = -14.0;
           if (acscore[q] > 19) acscore[q] = 19.0;
           acscore[q] = 5.0*(acscore[q]+14.0)/33.0;
           acscore[q] = 0.4*acscore[q] + 0.6*gt[*s][*(s+1)];
       }
   }
                                                                                                                                 
   return;
}

static void splice_Glimmer(uchar *in_seqx, int ls, int us, int le, int ue,
                           uchar *in_seqy, int ys, int ye,
                           double *gtscore, double *agscore,
                           double *ctscore, double *acscore, int ori)
{
   int p, q, i;
   uchar  *s,*t, ch;
                                                                                                                                 
   for (i=0, s=in_seqx+ls-MAX_SPAN-1; i<2*MAX_SPAN+us-ls+3; nsegmentL[i++] = spl_encode[*s++]);
   for (i=0, s=in_seqx+le-2-MAX_SPAN-1; i<2*MAX_SPAN+ue-le+3; nsegmentR[i++] = spl_encode[*s++]);
                                                                                                                                 
   /* Glimmer specific matrices */
   for (i=0, s=in_seqx+ls-MAX_SPAN-1; i<2*MAX_SPAN+us-ls+3; asegmentL[i++] = *s++);
   for (i=0, s=in_seqx+le-2-MAX_SPAN-1; i<2*MAX_SPAN+ue-le+3; asegmentR[i++] = *s++);
                                                                                                                                 
   if (ori==FWD || ori==BOTH) {

      for (p=0, s=nsegmentL+MAX_SPAN, t=asegmentL+MAX_SPAN; p<=us-ls+1; p++, s++, t++) {
          gtscore[p] = ScoreDonor_Glimmer(t-GLIMMER_XSPAN, Glimmer_TRAIN_DIR);
                                                                                                                                 
          if (gtscore[p] < 0) gtscore[p] = 0.0;
          if (gtscore[p] > 0.31) gtscore[p] = 0.31;
          gtscore[p] = 5.0*(gtscore[p]+0.0)/0.31;
          gtscore[p] = 0.2*gtscore[p] + 0.8*gt[*s][*(s+1)];
      }
      for (q=ue-le+1, s=nsegmentR+MAX_SPAN+ue-le+2, t=asegmentR+MAX_SPAN+ue-le+2; q>=0; q--, s--, t--) {
          agscore[q] = ScoreAcceptor_Glimmer(t-GLIMMER_ISPAN-1, Glimmer_TRAIN_DIR);
                                                                                                                                 
          if (agscore[q] < -0.16) agscore[q] = -0.16;
          if (agscore[q] > 0.23) agscore[q] = 0.23;
          agscore[q] = 5.0*(agscore[q]+0.16)/0.39;
          agscore[q] = 0.2*agscore[q] + 0.8*ag[*(s-1)][*s];
      }
   }
                                                                                                                                 
   if (ori==BWD || ori==BOTH) {

       /* reverse complement the nsegments, 0-3 alphabet */
       for (s=nsegmentL, t=nsegmentL+2*MAX_SPAN+us-ls+3-1; s<t; s++, t--)
           { ch = 3-(*s); *s = 3-(*t); *t = ch; }
       for (s=nsegmentR, t=nsegmentR+2*MAX_SPAN+ue-le+3-1; s<t; s++, t--)
           { ch = 3-(*s); *s = 3-(*t); *t = ch; }
                                                                                                                                 
                                                                                                                                 
       /* reverse complement the asegments, ACTG alphabet */
       for (s=asegmentL, t=asegmentL+2*MAX_SPAN+us-ls+3-1; s<t; s++, t--)
            { ch = rev_compl[*s]; *s = rev_compl[*t]; *t = ch; }
       for (s=asegmentR, t=asegmentR+2*MAX_SPAN+ue-le+3-1; s<t; s++, t--)
            { ch = rev_compl[*s]; *s = rev_compl[*t]; *t = ch; }
                                                                                                                                 
       for (p=0, s=nsegmentL+MAX_SPAN+us-ls+2, t=asegmentL+MAX_SPAN+us-ls+2; p<=us-ls+1; p++, s--, t--)  {
           ctscore[p] = ScoreAcceptor_Glimmer(t-GLIMMER_ISPAN-1, Glimmer_TRAIN_DIR);
                                                                                                                                 
           if (ctscore[p] < -0.16) ctscore[p] = -0.16;
           if (ctscore[p] > 0.23) ctscore[p] = 0.23;
           ctscore[p] = 5.0*(ctscore[p]+0.16)/0.39;
           ctscore[p] = 0.2*ctscore[p] + 0.8*ag[*(s-1)][*s];
       }
       for (q=ue-le+1, s=nsegmentR+MAX_SPAN, t=asegmentR+MAX_SPAN; q>=0; q--, s++, t++) {
           acscore[q] = ScoreDonor_Glimmer(t-GLIMMER_XSPAN, Glimmer_TRAIN_DIR);
                                                                                                                                 
           if (acscore[q] < 0) acscore[q] = 0.0;
           if (acscore[q] > 0.31) acscore[q] = 0.31;
           acscore[q] = 5.0*(acscore[q]+0.0)/0.31;
           acscore[q] = 0.2*acscore[q] + 0.8*gt[*s][*(s+1)];
       }
   }
                                                                                                                                 
   return;
}


splice_t *new_splice(char c, int xs, int xe, int ys, int ye, double score, splice_t *next)
{
   splice_t *sp = ckalloc(sizeof(splice_t));
   
   sp->type = c; sp->xs = xs; sp->xe = xe; 
   sp->ys = ys; sp->ye = ye; sp->score = score;
   sp->next = next;

   return sp; 
}

#ifdef DEBUG
static void print_splice(splice_t *g)
{
    printf("Type: %c  sx: %d  se: %d  ys: %d  score: %.3f\n",
            g->type, g->xs, g->xe, g->ys, g->score);

}
#endif

static void splice_donor(uchar *xseq, uchar *yseq, int M, int N, double *gt_score,
                         double *ct_score, double **max_Gf, double **max_Cf, 
                         double **start_Gi, double **start_Ci)
{
   int *CCf;
   double *mG, *mC, *sC, *sG;
   int *X;
   int i, j, tmp, ss, ssx, cx, c;
   uchar *s, *t;

   CCf = (int *)ckalloc((M+1)*sizeof(int));
   X = (int *)ckalloc((M+1)*sizeof(int));
   mG = *max_Gf = (double *)ckalloc((2*N+2)*sizeof(double));
   sG = *start_Gi = mG+(N+1);
   mC = *max_Cf = (double *)ckalloc((2*N+2)*sizeof(double));
   sC = *start_Ci = mC+(N+1);

   t = yseq; X[0] = CCf[0] = 0;
   for (j=1; j<=M; j++) { CCf[j] = j; X[j] = 0; } 
   
   mG[0] = mC[0] = -999999;
   for (j=0; j<=M; j++) {
      if ((100*gt_score[j])>mG[0]) { mG[0] = 100*gt_score[j]; sG[0] = j; }
      if ((100*ct_score[j])>mC[0]) { mC[0] = 100*ct_score[j]; sC[0] = j; } 
   }
   
   for (i=1; i<=N; i++, t++) {
     s = xseq;
     ss = CCf[0]; ssx = X[0];
     c = ++CCf[0]; cx = X[0];
     for (j=1; j<=M; j++, s++) {
          tmp=min(min(CCf[j]+1, ss+(*t!=*s)),c+1);
          if (tmp==c+1);
          else if (tmp==CCf[j]+1) cx = X[j];
          else cx = ssx + (*t==*s);
          c = tmp; ss = CCf[j]; CCf[j] = c; ssx = X[j]; X[j] = cx;
     }
   
     /* compute max_Gf and max_Cf */
     mG[i] = mC[i] = -999999; 
     for (j=0; j<=M; j++) {
        assert(X[j]+CCf[j]!=0);
        tmp = (int)(stepct(j)*X[j]/(double)(X[j]+CCf[j])*100); 
        if ((tmp+100*gt_score[j])>mG[i]) {
            mG[i] = tmp+100*gt_score[j]; sG[i] = j;
        }
        if ((tmp+100*ct_score[j])>mC[i]) {
            mC[i] = tmp+100*ct_score[j]; sC[i] = j;
        }
     } 
   } 
   free(CCf); free(X); 
}  

static void splice_donor_uni(uchar *xseq, uchar *yseq, int M, int N,
                             double *It_score, double **max_If, double **start_Ii)
{
   int *CCf, *X;
   double *mI, *sI;
   int i, j, tmp, ss, ssx, cx, c;
   uchar *s, *t;

   CCf = (int *)ckalloc((M+1)*sizeof(int));
   X = (int *)ckalloc((M+1)*sizeof(int));
   mI = *max_If = (double *)ckalloc((2*N+2)*sizeof(double));
   sI = *start_Ii = mI+(N+1);
  
   t = yseq; X[0] = CCf[0] = 0;
   for (j=1; j<=M; j++) { CCf[j] = j; X[j] = 0; }
 
   mI[0] = -999999;        
   for (j=0; j<=M; j++)                  
      if ((100*It_score[j])>mI[0]) { mI[0] = 100*It_score[j]; sI[0] = j; }      
 
   for (i=1; i<=N; i++, t++) {
     s = xseq;
     ss = CCf[0]; ssx = X[0];
     c = ++CCf[0]; cx = X[0];
     for (j=1; j<=M; j++, s++) {
          tmp=min(min(CCf[j]+1, ss+(*t!=*s)),c+1);
          if (tmp==c+1);
          else if (tmp==CCf[j]+1) cx = X[j];
          else cx = ssx + (*t==*s);
          c = tmp; ss = CCf[j]; CCf[j] = c; ssx = X[j]; X[j] = cx;
     }
  
     /* compute max_If */           
     mI[i] = -999999;        
     for (j=0; j<=M; j++) {
        assert(X[j]+CCf[j]!=0);
        tmp = (int)(stepct(j)*X[j]/(double)(X[j]+CCf[j])*100)+100*It_score[j];
        if (tmp>mI[i]) {
            mI[i] = tmp; sI[i] = j;
        }
     }
   }
   free(CCf); free(X);
}

     
static void splice_acceptor(uchar *xseq, uchar *yseq, int M, int N, 
                            double *ag_score, double *ac_score, double **max_Gb, 
                            double **max_Cb, double **end_Gi, double **end_Ci)
{
   int *CCb, *X; 
   double *mC, *mG, *eC, *eG;
   int tmp, i, j, ss, ssx, cx, c;
   uchar *t, *s;

   CCb = (int *)ckalloc((M+1)*sizeof(int));
   X = (int *)ckalloc((M+1)*sizeof(int));
   mG = *max_Gb = (double *)ckalloc((2*N+2)*sizeof(double));
   eG = *end_Gi = mG+(N+1);
   mC = *max_Cb = (double *)ckalloc((2*N+2)*sizeof(double));
   eC = *end_Ci = mC+(N+1);
   
   t = yseq+N-1; CCb[M] = X[M] = 0;
   for (j=M-1; j>=0; j--) { CCb[j] = M-j; X[j] = 0; }
   
   mG[N] = mC[N] = -999999;
   for (j=M; j>=0; j--) {
      if ((100*ag_score[j])>mG[N]) { mG[N] = 100*ag_score[j]; eG[N] = j+1; }
      if ((100*ac_score[j])>mC[N]) { mC[N] = 100*ac_score[j]; eC[N] = j+1; }
   }
   
   for (i=N-1; i>=0; i--, t--) {
     s = xseq+M-1; 
     ss = CCb[M]; ssx = X[M];
     c = ++CCb[M]; cx = X[M];
     for (j=M-1; j>=0; j--, s--) {
          tmp=min(min(CCb[j]+1, ss+(*t!=*s)),c+1);
          if (tmp==c+1) ; 
          else if (tmp==CCb[j]+1) cx = X[j];
          else cx = ssx + (*t==*s);
          c = tmp; ss = CCb[j]; CCb[j] = c; ssx = X[j]; X[j] = cx;
     }
     
     /* compute max_Gb and max_Cb */
     mG[i] = -999999; mC[i] = -999999;
     for (j=M; j>=0; j--) {
        assert(CCb[j]+X[j]!=0);
        tmp = (int)(stepct(M-j)*X[j]/(double)(CCb[j]+X[j])*100);
        if ((tmp+100*ag_score[j])>mG[i]) {
            mG[i] = tmp+100*ag_score[j]; eG[i] = j+1;
        }
        if ((tmp+100*ac_score[j])>mC[i]) {
            mC[i] = tmp+100*ac_score[j]; eC[i] = j+1;
        }
     } 
   } 
   free(CCb); free(X); 
}


static void splice_acceptor_uni(uchar *xseq, uchar *yseq, int M, int N,
                                double *aI_score, double **max_Ib, double **end_Ii)
{
   int *CCb, *X;
   double *mI, *eI;
   int tmp, i, j, ss, ssx, cx, c;
   uchar *t, *s;

   CCb = (int *)ckalloc((M+1)*sizeof(int));
   X = (int *)ckalloc((M+1)*sizeof(int));
   mI = *max_Ib = (double *)ckalloc((2*N+2)*sizeof(double));
   eI = *end_Ii = mI+(N+1);

   t = yseq+N-1; CCb[M] = X[M] = 0;
   for (j=M-1; j>=0; j--) { CCb[j] = M-j; X[j] = 0; }
  
   mI[N] = -999999;        
   for (j=M; j>=0; j--)  
      if ((100*aI_score[j])>mI[N]) { mI[N] = 100*aI_score[j]; eI[N] = j+1; }
 
   for (i=N-1; i>=0; i--, t--) {
     s = xseq+M-1;
     ss = CCb[M]; ssx = X[M];
     c = ++CCb[M]; cx = X[M];
     for (j=M-1; j>=0; j--, s--) {
          tmp=min(min(CCb[j]+1, ss+(*t!=*s)),c+1);
          if (tmp==c+1) ;
          else if (tmp==CCb[j]+1) cx = X[j];
          else cx = ssx + (*t==*s);
 
          c = tmp; ss = CCb[j]; CCb[j] = c; ssx = X[j]; X[j] = cx;
     }
       
     /* compute max_Ib */           
     mI[i] = -999999; 
     for (j=M; j>=0; j--) {
        assert(CCb[j]+X[j]!=0);
        tmp = (int)(stepct(M-j)*X[j]/(double)(CCb[j]+X[j])*100)+100*aI_score[j];
        if (tmp>mI[i]) {
            mI[i] = tmp; eI[i] = j+1;
        }
     }
   }
   free(CCb); free(X);
}


static int stepct(int n)
{
  if (n<0) fatal("splice.c: Negative value in stepct().");
  if (n<=4) return 9;     
  if (n<=8) return 10;
  if (n<=12) return 12;
  return 12;
}

void splice_close()
{
   free(nsegmentL);
   free(nsegmentR);
   free(asegmentL);
   free(asegmentR);
   UnLoadSites_GeneSplicer();

   spliceInit = 0;
}

